/***************************************************************************
  **************************************************************************
  
                SOFT: SO(3) Fourier transform code
  
   Peter Kostelec, Dan Rockmore
   {geelong,rockmore}@cs.dartmouth.edu
  
   Contact: Peter Kostelec
            geelong@cs.dartmouth.edu
  
  
   Copyright 2003 Peter Kostelec, Dan Rockmore
  
  
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
  
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
  
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  
   Commercial use is absolutely prohibited.
  
   See the accompanying LICENSE file for details.
  
  ************************************************************************
  ************************************************************************/



/***********************************************************************

  some "primitive" utility functions:

  vec_add() - add two vectors together
  vec_mul() - multiply vector by a scalar
  vec_pt_mul() - pointwise-multiply two vectors
  vec_inner() - inner (i.e. dot) product of two vectors
  transpose() - transpose an n x m matrix (assumed to be stored
                in a 1-D array in ROW-MAJOR format)
  totalCoeffs() - total number of coefficients calculated in a
                  full SO3 forward transform
  numCoeffs() - in a full forward SO3 transform, how many coefficients
                are computed for order m1 and bandwidth bw ?
  howMany() - how many coeffs computed at orders m1, m2 and bandwidth bw?
  


************************************************************************/

#include "utils.h"

/************************************************************************/
/* vector arithmetic operations */
/************************************************************************/
/* does result = data1 + data2 */
/* result and data are vectors of length n */

void vec_add( double *data1,
	      double *data2,
	      double *result,
	      int n )
{
  int k ;

  for (k = 0; k < n % 4; ++k)
    result[k] = data1[k] + data2[k];

  for ( ; k < n ; k += 4)
    {
      result[k] = data1[k] + data2[k];
      result[k + 1] = data1[k + 1] + data2[k + 1];
      result[k + 2] = data1[k + 2] + data2[k + 2];
      result[k + 3] = data1[k + 3] + data2[k + 3];
    }
}
/************************************************************************/
/************************************************************************/
/*
   vec_mul(scalar,data1,result,n) multiplies the vector 'data1' by
   'scalar' and returns in result 
*/
void vec_mul( double scalar,
	      double *data1,
	      double *result,
	      int n)
{
   int k;

   for( k = 0; k < n % 4; ++k)
     result[k] = scalar * data1[k];

   for( ; k < n; k +=4)
     {
       result[k] = scalar * data1[k];
       result[k + 1] = scalar * data1[k + 1];
       result[k + 2] = scalar * data1[k + 2];
       result[k + 3] = scalar * data1[k + 3];
     }

}

/************************************************************************/
/************************************************************************/

/* point-by-point multiplication of length n vectors data1, data2,
 places result in result */

void vec_pt_mul( double *data1,
		 double *data2,
		 double *result,
		 int n)
{
  int k;
  
  for(k = 0; k < n % 4; ++k)
    result[k] = data1[k] * data2[k];
  
  for( ; k < n; k +=4)
    {
      result[k] = data1[k] * data2[k];
      result[k + 1] = data1[k + 1] * data2[k + 1];
      result[k + 2] = data1[k + 2] * data2[k + 2];
      result[k + 3] = data1[k + 3] * data2[k + 3];
    }
 
}

/**********************************************************/
/**********************************************************/
/*
  compute the inner (dot) product of two vectors data1
  and data2, each of length n. The result will be RETURNED
*/

double vec_inner( double *data1,
		  double *data2,
		  int n )
{
  int k ;
  double tmp0, tmp1, tmp2, tmp3 ;

  tmp0 = 0.0 ;
  tmp1 = 0.0 ;
  tmp2 = 0.0 ;
  tmp3 = 0.0 ;
  

  for ( k = 0 ; k < n % 4 ; ++k )
    tmp0 += data1[ k ] * data2[ k ] ;

  for ( ; k < n ; k += 4 )
    {
      tmp0 += data1[ k ] * data2[ k ] ;
      tmp1 += data1[ k + 1] * data2[ k + 1] ;
      tmp2 += data1[ k + 2] * data2[ k + 2] ;
      tmp3 += data1[ k + 3] * data2[ k + 3] ;
    }
  
  return ( tmp0 + tmp1 + tmp2 + tmp3 ) ;
}

/*******************************************************/
/*******************************************************/
/*
  transpose() - transpose an m x n matrix (assumed to be stored
                in a 1-D array in ROW-MAJOR format)

  This function won't be efficient, but it oughta work.

 arguments: arrayIn: 1-D input array of length m*n,
                     treated as an m x n matrix
	    
	    arrayOut: 1-D output array of length m*n,
                      the TRANSPOSE of arrayIn

*/

void transpose( double *arrayIn,
		double *arrayOut,
		int m,
		int n )
{
  int j, k;
  double *inPtr, *outPtr;

  inPtr = arrayIn ;
  outPtr = arrayOut ;

  for( j = 0 ; j < n ; j ++ )
    for( k = j ; k < m*n ; k += n )
      *outPtr++ = inPtr[ k ] ;

}

/***********************************************************/
/*

  What follows are some "indexing/quantity" type functions
  useful when doing the full transform

*/

/*
  At bandwidth bw, what are the TOTAL NUMBER of coefficients
  that will be computed when doing a full SO3 forward transform ?

  total number = 1/3 * bw * (4 * bw^2 - 1 )

*/

int totalCoeffs( int bw )
{

  /* note that 4 * bw^3 - bw is always a multiple of 3,
     so integer division won't be messy */

  return ( (4 * bw * bw * bw - bw ) / 3 ) ;
}

/***************************************************/

/*
  At bandwidth bw and order m1, what's the total number of
  coefficients that will be computed, when doing a full SO3
  forward transform, at this order?

  That is, fix m1. At each "legal" m2, I will compute so many
  coefficients. Let me add all those numbers up.

  total number at order m1, for bandwidth bw: bw^2 - m1^2

*/

int numCoeffs( int bw,
	       int m1 )
{

  return ( bw*bw - m1*m1 ) ;

}

/************************************************/
/*
  At orders m1, m2, and bandwidth bw, how many coefficients
  will be computed ?

  let m = Max( |m1|, |m2| )

  The number of coefficients is = bw - m

*/

int howMany( int m1,
	     int m2,
	     int bw )
{
  int m ;

  m = MAX( ABS( m1 ) , ABS( m2 ) );

  return ( bw - m ) ;

}
